# -*- coding: utf-8 -*-

from datetime import datetime
import ephem
from typing import List
import pandas as pd

TYPE_ASTRONOMICAL = "astronomical"

STATE_SPRING = 0
STATE_SUMMER = 1
STATE_FALL = 2
STATE_WINTER = 3

NORTHERN = "north"
SOUTHERN = "south"

HEMISPHERE_SEASON_SWAP = {
    STATE_SPRING: STATE_FALL,
    STATE_SUMMER: STATE_WINTER,
    STATE_FALL: STATE_SPRING,
    STATE_WINTER: STATE_SUMMER
}

def get_season(date, hemisphere: str = "north", labels: List[str] = ["Spring", "Summer", "Fall", "Winter"], tracking_type: str = "astronomical") -> str:
    """
    Return the season of the given date depending on the latitude and location.

    Args:
        date: datetime object or pandas Series with datetime objects
        hemisphere: "north" or "south", default is "north"
        labels: array of season names, default is ["Spring", "Summer", "Fall", "Winter"]
        tracking_type: Type of season definition. Options are "meteorological" or "astronomical". Default is "astronomical"
        
    Returns:
        A string with season name or a pandas series with strings, depending on the input
    """
    if isinstance(date, pd.Series):
        return date.apply(lambda x: get_season(x, hemisphere, labels, tracking_type))
    
    if tracking_type == TYPE_ASTRONOMICAL:
        spring_start = ephem.next_equinox(str(date.year)).datetime()
        summer_start = ephem.next_solstice(str(date.year)).datetime()
        autumn_start = ephem.next_equinox(spring_start).datetime()
        winter_start = ephem.next_solstice(summer_start).datetime()
    else:
        spring_start = datetime(date.year, 3, 1)
        summer_start = datetime(date.year, 6, 1)
        autumn_start = datetime(date.year, 9, 1)
        winter_start = datetime(date.year, 12, 1)
    
    if spring_start <= date < summer_start:
        season = STATE_SPRING
    elif summer_start <= date < autumn_start:
        season = STATE_SUMMER
    elif autumn_start <= date < winter_start:
        season = STATE_FALL
    elif winter_start <= date or spring_start > date:
        season = STATE_WINTER
    
    # Swap the season if on southern hemisphere
    if hemisphere == SOUTHERN:
        return labels[HEMISPHERE_SEASON_SWAP.get(season, season)]
    return labels[season]



# # -*- coding: utf-8 -*-

import unittest
from pyedautils.geopy import get_altitude_lat_long, get_lat_long_address, get_altitude_lv95, convert_wsg84_to_lv95, GeocodingError, get_coordindates_ch_plz, get_distance_between_two_points

class TestGeopyhelpers(unittest.TestCase):
    """Tests for "geopyhelpers.py"."""
    
    def test_get_altitude_lat_long(self):
        """Test get_altitude() with different lat/long pairs"""
        self.assertIn(round(get_altitude_lat_long(47.01450, 8.30620),0), range(438,445))
        with self.assertRaises(ValueError):
            get_altitude_lat_long(57.01450, 8.30620)
        with self.assertRaises(ValueError):
            get_altitude_lat_long(100, 10)
        with self.assertRaises(ValueError):
            get_altitude_lat_long(-100, 10)
        with self.assertRaises(ValueError):
            get_altitude_lat_long(10, -200)
        with self.assertRaises(ValueError):
            get_altitude_lat_long(10, 200)

    def test_get_lat_long_address(self):
        """Test get_lat_long() with different addresses"""
        self.assertEqual(round(get_lat_long_address("Technikumstrasse 21, 6048 Horw")[0], 1), 47.0)
        self.assertEqual(round(get_lat_long_address("Technikumstrasse 21, 6048 Horw")[1], 1), 8.3)
        with self.assertRaises(GeocodingError):
            get_lat_long_address("Highway to hell, Horw, Switzerland")        
        
    def test_convert_wsg84_to_lv95(self):
        """Test convert_wsg84_to_lv95() with different coodrinates"""
        self.assertIn(round(convert_wsg84_to_lv95(47.01331, 8.30612)[0], 0), range(2665959,2665962))
        self.assertIn(round(convert_wsg84_to_lv95(47.01450, 8.30620)[1], 0), range(1207300,1207450))
        with self.assertRaises(GeocodingError):
            convert_wsg84_to_lv95(57.01450, 8.30620)
        with self.assertRaises(GeocodingError):
            convert_wsg84_to_lv95(47.01450, 15.30620)
    
    def test_get_altitude_lv95(self):
        """Test get_altitude_ch_lv95() with different lat/long pairs"""
        self.assertIn(round(get_altitude_lv95([2665949.0,1207341.8]),0), range(438,445))
        self.assertIn(round(get_altitude_lv95(convert_wsg84_to_lv95(47.01450, 8.30620)),0), range(438,445))
        
        
    def test_find_coordindates_ch_plz(self):
        self.assertAlmostEqual(get_coordindates_ch_plz(6048)[0], 47.01, delta=0.1)
        self.assertAlmostEqual(get_coordindates_ch_plz(6048)[1], 8.3, delta=0.1)
        with self.assertRaises(GeocodingError):
            get_coordindates_ch_plz(424242)
            
    def test_get_distance_between_two_points(self):
        self.assertEqual(round(get_distance_between_two_points(get_coordindates_ch_plz(6048), get_coordindates_ch_plz(3800)), 0), 50)
        self.assertAlmostEqual(get_distance_between_two_points(get_coordindates_ch_plz(6048), get_lat_long_address("20 W 34th, New York")), 6323.833, delta = 10)
        with self.assertRaises(GeocodingError):
            get_distance_between_two_points(get_coordindates_ch_plz(5555555), get_coordindates_ch_plz(3800))
            
if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import pandas as pd
from pyedautils.geopy import get_distance_between_two_points

def get_current_station_data():
    """
    Gets current measurement data of all Meteo Swiss stations.

    Returns:
        df: pandas dataframe with station data
    """
    try:
        endpoint = "https://data.geo.admin.ch/ch.meteoschweiz.messnetz-automatisch/ch.meteoschweiz.messnetz-automatisch_de.csv"
        data = pd.read_csv(endpoint, encoding= 'unicode_escape', sep = ";")
        df = data[data['Messungen'].notna()]
           
        return df
    except:
        return pd.DataFrame()

def find_nearest_station(lat: float, long: float, altitude:float, sensor=None):
    """
    Returns station id of closest meteo swiss station to a coordinate.

    Args:
        lat (float): Latitude in decimal degrees.
        long (float): Longitude in decimal degrees.
        altitude (float): Altitude in meters above sea level.
        sensor (str): temp, globrad, relhum or rain  

    Returns:
        str: Meteo Swiss station ID
    """
    
    switzerland_lat_min = 45.67
    switzerland_lat_max = 47.92
    switzerland_long_min = 5.7
    switzerland_long_max = 10.7
    
    if(not((switzerland_lat_min <= lat <= switzerland_lat_max) and (switzerland_long_min <= long <= switzerland_long_max))):
        raise ValueError("Coordinates not in range for Swiss coordinate system")
    
    all_station_data = get_current_station_data()
    
    lats=all_station_data["Breitengrad"].astype("float")
    lons=all_station_data["Längengrad"].astype("float")
    
    distances=pd.DataFrame([get_distance_between_two_points((lat,long),(lats[i],lons[i])) for i in range(len(lats)) ])
    distances.columns = ['value']
    distances = distances.sort_values(by='value')
    distances = distances.reset_index()

    i = 0
    
    while i < len(distances):
        id = distances.loc[i,"index"]
        sensors = all_station_data.loc[id,"Messungen"]        
        found = False
        stationID = all_station_data.loc[id,"Abk."]
        stationIDString = all_station_data.loc[id,"Station"]
        stationAltitude = float(all_station_data.loc[id,"Stationshöhe m ü. M."])
        if(sensor==None):
            break
        else:
            if((float(altitude) > (stationAltitude - 150.0)) and (float(altitude) < (stationAltitude + 150.0))):
                if(sensor=="temp" and ("Temperatur" in sensors)):
                    print("Closest station for temp: " + str(stationIDString))
                    found = True
                    break
                elif(sensor=="globrad" and ("Globalstrahlung" in sensors)):
                    print("Closest station for globalrad: " + str(stationIDString))
                    found = True
                    break
                elif(sensor=="relhum" and ("Feuchte" in sensors)):
                    print("Closest station for relhum: " + str(stationIDString))
                    found = True
                    break
                elif(sensor=="rain" and ("Niederschlag" in sensors)):
                    print("Closest station for rain: " + str(stationIDString))
                    found = True
                    break
                else:
                    i += 1
                    continue
            else:
                i += 1
                continue
            
            if(found):
                break
    return stationID




from geopy.geocoders import Nominatim
import pgeocode
import requests
import pandas as pd
import time
from typing import List, Union, Tuple
import math

class GeocodingError(Exception):
    pass

def convert_wsg84_to_lv95(lat: float, long: float) -> List[float]:
    """
    Converts WGS84 latitude and longitude coordinates to Swiss coordinate system LV95.

    Args:
        lat (float): Latitude in decimal degrees.
        long (float): Longitude in decimal degrees.

    Returns:
        List[float]: A list containing x and y coordinates in LV95 system, e.g., [xcoord, ycoord].
    """
    if not (5.2 <= long <= 11) or not (45.4 <= lat <= 48.2):
        raise GeocodingError("Coordinates not in range for Swiss coordinate system LV95")

    latitude = str(lat)
    longitude = str(long)
    query = f'http://geodesy.geo.admin.ch/reframe/wgs84tolv95?easting={longitude}&northing={latitude}'
    try:
        r = requests.get(query).json()
        coord_list_lv95 = r.get("coordinates", [])
        return coord_list_lv95
    except Exception as e:
        raise GeocodingError(f"Failed to convert WGS84 to LV95: {e}") from e

def get_altitude_lv95(coord_list_lv95: List[float]) -> float:
    """
    Returns altitude in meters above sea level for the given LV95 coordinates.
    The geo.admin.ch api gets used.

    Args:
        coord_list_lv95 (List[float]): LV95 coordinates as [xcoord, ycoord].

    Returns:
        float: Altitude in meters above sea level.
    """
    query = f'https://api3.geo.admin.ch/rest/services/height?easting={coord_list_lv95[0]}&northing={coord_list_lv95[1]}'
    try:       
        r = requests.get(query).json()
        altitude = float(r.get("height", 0))
        return altitude
    except Exception as e:
        raise GeocodingError(f"Failed to get altitude for LV95 coordinates, Error: {e}") from e

def get_altitude_lat_long(lat: float, long: float) -> float:
    """
    Returns altitude in meters above sea level for the given WGS84 coordinates.
    The opentopodata.org api gets used.

    Args:
        lat (float): Latitude in decimal degrees.
        long (float): Longitude in decimal degrees.

    Returns:
        float: Altitude in meters above sea level.
    """
    switzerland_lat_min = 45.67
    switzerland_lat_max = 47.92
    switzerland_long_min = 5.7
    switzerland_long_max = 10.7
    
    if(not((switzerland_lat_min <= lat <= switzerland_lat_max) and (switzerland_long_min <= long <= switzerland_long_max))):
        raise ValueError("Coordinates not in range for Swiss coordinate system")
    
    latitude = str(lat)
    longitude = str(long)
    query = f'https://api.opentopodata.org/v1/eudem25m?locations={latitude},{longitude}'
    try:
        r = requests.get(query).json()
        if(pd.json_normalize(r, 'results')['elevation'][0] == None):
            altitude = 0
        else:
            altitude = float(pd.json_normalize(r, 'results')['elevation'].values[0])
        time.sleep(1)  # API rate limiting
        return round(altitude,1)
    except Exception as e:
        raise GeocodingError(f"Failed to get altitude for WGS84 coordinates: {e}") from e

def get_lat_long_address(address: str) -> Union[List[float], None]:
    """
    Returns latitude and longitude coordinates for the given address.

    Args:
        address (str): The address to geocode.

    Returns:
        Union[List[float], None]: A list containing latitude and longitude coordinates, e.g., [latitude, longitude], or None if the address could not be geocoded.
    """
    try:
        nom = Nominatim(user_agent="myPythonScript")
        for _ in range(3):
            n = nom.geocode(address)
            if n is not None:
                break
            time.sleep(1)  # API rate limiting
        else:
            raise GeocodingError("Failed to geocode address after 3 attempts")

        return [n.latitude, n.longitude]
    except Exception as e:
        raise GeocodingError(f"Failed to geocode address, Error: {e}") from e

def get_coordindates_ch_plz(plz: int) -> Tuple[float, float]:
    """
    Returns latitude and longitude for a Swiss postal code.

    Args:
        plz (int): Postal Code (Postleitzahl)

    Returns:
        tuple (lat: float, lon: float)
    """
    try:
        nomi = pgeocode.Nominatim('ch') 
        data=nomi.query_postal_code(plz)
        coordinates = data.latitude.astype(float), data.longitude.astype(float)
        if(str(coordinates[0]) == "nan"):
            raise GeocodingError(f"Failed to get lat/long from plz {plz}")
    except Exception:
        raise GeocodingError(f"Failed to get lat/long for plz {plz}")
     
    return coordinates

def get_distance_between_two_points(coord1, coord2):
    """
    Calculates the distance between two points on the Earth's surface given their latitude and longitude coordinates.
    
    Args:
        coord1 (tuple): Latitude and longitude of the first point in degrees, as a tuple (lat1, lon1).
        coord2 (tuple): Latitude and longitude of the second point in degrees, as a tuple (lat2, lon2).
        
    Returns:
        float: Distance between the two points in kilometers.
    
    """
    # Radius of Earth in km
    R = 6373.0
    
    # Convert latitude and longitude from degrees to radians
    lat1, lon1 = math.radians(coord1[0]), math.radians(coord1[1])
    lat2, lon2 = math.radians(coord2[0]), math.radians(coord2[1])
    
    # Calculate differences in longitude and latitude
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    
    # Apply Haversine formula to calculate distance
    a = math.sin(dlat / 2)**2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2)**2
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    distance = R * c
    
    return round(distance, 3)


# # -*- coding: utf-8 -*-

import unittest
from pyedautils.meteo_swiss import get_current_station_data, find_nearest_station
from pyedautils.geopy import get_altitude_lat_long, get_coordindates_ch_plz

class TestGeopyhelpers(unittest.TestCase):
    """Tests for "geopyhelpers.py"."""
    
    def test_get_current_station_data(self):
        """Test get_current_station_data()"""
        self.assertGreaterEqual(get_current_station_data().shape[0], 1)
        
    def test_find_nearest_station(self):
        coord = get_coordindates_ch_plz(6048)
        altitude = get_altitude_lat_long(coord[0], coord[1])
        self.assertEqual(find_nearest_station(coord[0], coord[1], altitude, sensor="temp"),"LUZ")
        coord = get_coordindates_ch_plz(6197)
        altitude = get_altitude_lat_long(coord[0], coord[1])
        self.assertEqual(find_nearest_station(coord[0], coord[1], altitude, sensor="temp"), "FLU")
        self.assertEqual(find_nearest_station(coord[0], coord[1], altitude, sensor="relhum"), "FLU")
        self.assertEqual(find_nearest_station(coord[0], coord[1], altitude, sensor="globrad"), "BAN")
        self.assertEqual(find_nearest_station(coord[0], coord[1], altitude, sensor="rain"), "FLU")
        
if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import unittest
import pandas as pd
from pandas.testing import assert_series_equal
from datetime import datetime

from pyedautils.season import get_season

class TestSeasonFunction(unittest.TestCase):
    def test_season_northern_hemisphere_astronomical(self):
        # Test dates for the northern hemisphere
        test_cases = [
            (datetime(2024, 3, 20, 3, 5), "Winter"),
            (datetime(2024, 3, 20, 3, 7), "Spring"),
            (datetime(2024, 6, 20, 20, 50), "Spring"),
            (datetime(2024, 6, 20, 21, 52), "Summer"),
            (datetime(2024, 9, 22, 12, 42), "Summer"),
            (datetime(2024, 9, 22, 12, 44), "Fall"),
            (datetime(2024, 12, 21, 9, 19), "Fall"),
            (datetime(2024, 12, 21, 9, 21), "Winter"),
        ]
        for date, expected_season in test_cases:
            with self.subTest(date):
                self.assertEqual(get_season(date=date), expected_season)

    def test_season_northern_hemisphere_meteorological(self):
        # Test dates for the northern hemisphere with meteorological datechanges
        test_cases = [
            (datetime(2024, 2, 27), "Winter"),
            (datetime(2024, 3, 1), "Spring"),
            (datetime(2024, 5, 30), "Spring"),
            (datetime(2024, 6, 1), "Summer"),
            (datetime(2024, 8, 31), "Summer"),
            (datetime(2024, 9, 1), "Fall"),
            (datetime(2024, 11, 30), "Fall"),
            (datetime(2024, 12, 1), "Winter"),
        ]
        for date, expected_season in test_cases:
            with self.subTest(date):
                self.assertEqual(get_season(date=date, tracking_type="meteorological"), expected_season)
                
    def test_season_southern_hemisphere_astronomical(self):
        # Test dates for the southern hemisphere
        test_cases = [
            (datetime(2024, 3, 23), "Fall"),  # Spring in the Northern Hemisphere
            (datetime(2024, 6, 30), "Winter"),  # Summer in the Northern Hemisphere
            (datetime(2024, 9, 24), "Spring"),  # Fall in the Northern Hemisphere
            (datetime(2024, 12, 24), "Summer"),  # Winter in the Northern Hemisphere
        ]
        for date, expected_season in test_cases:
            with self.subTest(date):
                self.assertEqual(get_season(date=date, hemisphere="south"), expected_season)

    def test_custom_labels(self):
        # Test with custom labels
        custom_labels = ["frühling", "sommer", "herbst", "winter"]
        test_cases = [
            (datetime(2024, 3, 20, 3, 5), "winter"),
            (datetime(2024, 3, 20, 3, 7), "frühling"),
            (datetime(2024, 6, 20, 20, 50), "frühling"),
            (datetime(2024, 6, 20, 21, 52), "sommer"),
            (datetime(2024, 9, 22, 12, 42), "sommer"),
            (datetime(2024, 9, 22, 12, 44), "herbst"),
            (datetime(2024, 12, 21, 9, 19), "herbst"),
            (datetime(2024, 12, 21, 9, 21), "winter"),
        ]
        for date, expected_season in test_cases:
            with self.subTest(date):
                self.assertEqual(get_season(date=date, labels=custom_labels), expected_season)
       
    def test_pandas_series(self):
        # Test passing a pandas Series object
        pd_datetime_series = pd.Series([datetime(2024, 3, 31, 3, 5), datetime(2024, 6, 22), datetime(2024, 9, 24), datetime(2024, 12, 24)])
        expected_seasons = pd.Series(["Spring", "Summer", "Fall", "Winter"])
        assert_series_equal(get_season(date=pd_datetime_series), expected_seasons)
                
if __name__ == '__main__':
    unittest.main()
